// --8<--8<--8<--8<--
//
// Copyright (C) 2013 Smithsonian Astrophysical Observatory
//
// This file is part of transform_ray
//
// transform_ray is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or (at
// your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// -->8-->8-->8-->8--

#include <algorithm>
#include <vector>
#include <iostream>
#include <iterator>

#include <Eigen/Geometry>
#include <boost/lexical_cast.hpp>

#include <pcrecpp.h>

#include <cxcparamx/simple.h>
#include <suplibxx/str.h>

#include <suplib/debug.h>

#include <tracefctxx/TraceFct.h>

#include "param.h"

using namespace Eigen;


Pars::Pars( int argc, char** argv ) {

    TraceFct tf( "read parameters" );

    // stolen from Perl's Regexp::Common module
    string re_real_bare = "(?:(?i)(?:[-+]?)(?:(?=[.]?[0123456789])(?:[0123456789]*)(?:(?:[.])(?:[0123456789]{0,}))?)(?:(?:[E])(?:(?:[-+]?)(?:[0123456789]+))|))";

    //
    string re_real = "\\s*(\\s*" + re_real_bare + "\\s*)\\s*";

    // match w, x, y, z
    pcrecpp::RE re_quat(   re_real + ","
			 + re_real + ","
			 + re_real + ","
			 + re_real
		      );

    // match @angle, x, y, z
    pcrecpp::RE re_aangle( "@"
			 + re_real + ","
			 + re_real + ","
			 + re_real + ","
			 + re_real
		      );

    string buffer;

    // Eigen defaults to column major order, but dvm3_RotMat is in
    // row-major order
    Transform<double,3,Affine,RowMajor> etx;

    cxcparamx::simple pf( argc, argv );

    pf.get( "version", version );
    pf.get( "help", help );

    if ( version || help )
	return;

    debug_init( pf.getstr( "debug" ).c_str( ) );

    pf.get( "input", input );
    pf.get( "output", output );

    // transform parameter may be a Quaternion, Axis-Angle, 3x3 rotation matrix,
    // or 4x4 augmented/affine/homogeneous matrix

    pf.get( "coeff", buffer );

    double c0, c1, c2, c3;

    bool found = false;

    if ( buffer == "identity" ) {

	etx = Quaterniond( 0, 0, 0, 0 );
	found = true;
    }

    else if ( re_quat.FullMatch( buffer , &c0, &c1, &c2, &c3 ) ) {

	etx = Quaterniond( c0, c1, c2, c3 );
	found = true;

    }

    else if ( re_aangle.FullMatch( buffer , &c0, &c1, &c2, &c3 ) ) {

	etx = AngleAxisd( c0, Vector3d( c1, c2, c3 ).normalized() );
	found = true;

    }

    else {

	vector<string> scoeff;
	suplib::tok( scoeff, buffer, "\t\n ,}{()", true );

	vector<double> coeff;

	try {
	    std::transform( scoeff.begin(), scoeff.end(),
			    back_inserter(coeff),
			    boost::lexical_cast<double,string> );

	    switch( coeff.size() ) {

	    case 9:	// 3X3 Rotation Matrix
		etx = Map<Matrix<double,3,3,RowMajor> >(&coeff[0]);
		found = true;
		break;

	    case 16:    // 4x4 Affine Matrix
		etx = Map<Matrix<double,4,4,RowMajor> >(&coeff[0]);
		found = true;
		break;

	    }
	} catch ( ... ) { }

    }

    if ( ! found )
	tf.die( "cannot parse the 'coeff' parameter into something recognizeable\n" );

    // translation
    double x, y, z;
    pf.get( "x", x );
    pf.get( "y", y );
    pf.get( "z", z );

    pf.get( "translate", buffer );

    // apply translation before or after transformation, remember that
    // we're right multiplying vectors, so it's Tn * ... * T1 * T0 * v

    // apply translation before transformation
    // so it's Tn * ... * T1 * T0 * v
    if ( buffer == "before" || buffer == "pre" ) {

	etx *= Translation3d( x, y, z );

    }

    // apply translation after transformation
    else if ( buffer == "after" || buffer == "post" ) {

	etx = Translation3d( x, y, z ) * etx;

    }

    // do nothing
    else if ( buffer == "ignore" || buffer == "none" || buffer == "" ) {

    }

    // whine
    else {

	tf.die( "illegal value for translate parameter\n" );

    }

    if ( pf.getb( "inverse") )
	etx = etx.inverse();

    Matrix<double, 3, 3, RowMajor>::Map(rot[0]) = etx.rotation();

    Vector3d eoff( etx.translation() );

    off = dvm3_Vector( eoff[0], eoff[1], eoff[2] );

    if ( dbflag( "xfrm" ) ) {
	cerr << "Rot:\n" << etx.rotation() << endl;
	cerr << "Off:\n" << eoff.transpose() << endl;
    }

}
