// --8<--8<--8<--8<--
//
// Copyright (C) 2013 Smithsonian Astrophysical Observatory
//
// This file is part of transform_ray
//
// transform_ray is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or (at
// your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// -->8-->8-->8-->8--

// rot_BPipe.cc

/****************************************************************************
 * Author     : Terry Gaetz
 *
 * Description: handle bpipe stuff.
 */

#include <tracefctxx/TraceFct.h>

#include "rot_BPipe.h"



/* Routine to get the data pointer for a named data packet field */

static char*
dpktf_data( BPipe* bpipe, const char *name, void *data ) {

  TraceFct tf("dpktf_data");

  DpktField *field = bpipe_dpktf(bpipe, name);

  if (field == NULL)
      tf.die( "%s isn't in input data stream", name);

  return bpipe_dpktf_data(field, data);
}

//=========================================================================
// ctors/dtors...

//-------------------------------------------------------------------------
rot_BPipe::
~rot_BPipe()
{
  TraceFct tf("~rot_BPipe()");

  if ( bpipe_errno != BPNOERROR )
      tf.die( bpipe_strerror(bpipe_errno) );

  bpipe_delete(bpipe);
}

//-------------------------------------------------------------------------
rot_BPipe::
rot_BPipe( string const& instream,
           string const& outstream
        )
  : n_ray(0),
    have_polstate(0)
{
  TraceFct tf("rot_BPipe(...)");

  init( instream, outstream );
  bpipe_errno = BPNOERROR;
}

//-------------------------------------------------------------------------
void rot_BPipe::
init( const string& instream, const string& outstream )
{
  TraceFct tf("rot_BPipe::init");

  /*-------------------------------------------------------------------
   * acquire & set up input & output bpipes;
   * set up pointers to data fields - attach BPipe fields to ray struct...
   */

  /* Setup bpipe input and output streams */

  if ((bpipe = bpipe_new()) == NULL)
      tf.die( "couldn't open bpipe: %s", bpipe_strerror(bpipe_errno));


  if ( bpipe_input(bpipe, instream.c_str() ) )
      tf.die( "couldn't open input file `%s': %s", instream.c_str(), bpipe_strerror(bpipe_errno));

  if (NULL == (bpo = bpipe_output(bpipe, outstream.c_str() )))
      tf.die( "couldn't open input file `%s': %s", outstream.c_str(), bpipe_strerror(bpipe_errno));

  if ( NULL == (data = bpipe_map_alloc(bpipe, 1, NULL)) )
      tf.die( "unable to allocate memory for data packet: %s", bpipe_strerror(bpipe_errno));

  //-----------------------------------
  // required fields...

  raydata.pos     = (DVector3*) dpktf_data(bpipe, BP_AXAF_phot_pos,    data);
  raydata.dir     = (DVector3*) dpktf_data(bpipe, BP_AXAF_phot_dir,    data);
  raydata.energy  = (double*)   dpktf_data(bpipe, BP_AXAF_phot_energy, data);

  //-----------------------------------
  // optional fields...

  DpktField* cosfield = bpipe_dpktf( bpipe, BP_AXAF_pol_cos );
  DpktField* sinfield = bpipe_dpktf( bpipe, BP_AXAF_pol_cos );
  if ( cosfield != NULL && sinfield != NULL ) {

      have_polstate = 1;
      raydata.c2 = (DCVector2*) bpipe_dpktf_data( cosfield, data );
      raydata.s2 = (DCVector2*) bpipe_dpktf_data( sinfield, data );

  }

  //-----------------------------------
  // write out the header...

  if (bpipe_write_hdr(bpipe))
      tf.die( "bpipe_write_hdr: %s", bpipe_strerror(bpipe_errno));

}

//=========================================================================
// i/o...

//-------------------------------------------------------------------------
int rot_BPipe::
read(  rl_Ray& ray )
{

  // Process the ray if we have one.

  size_t rc = bpipe_read_dpkts(bpipe, data, 1);

  if ( rc )
  {
    // copy ray/surface data over into ray and surface...

    ray.position(0)  =  raydata.pos->x;
    ray.position(1)  =  raydata.pos->y;
    ray.position(2)  =  raydata.pos->z;
    ray.direction(0) =  raydata.dir->x;
    ray.direction(1) =  raydata.dir->y;
    ray.direction(2) =  raydata.dir->z;
    ray.energy()     = *raydata.energy;

    if ( have_polstate )
    {
      rl_PolCSPOD cs;
      cs.c2_[0] = rl_Traits::complex( raydata.c2->q1.r, raydata.c2->q1.i );
      cs.c2_[1] = rl_Traits::complex( raydata.c2->q2.r, raydata.c2->q2.i );
      cs.s2_[0] = rl_Traits::complex( raydata.s2->q1.r, raydata.s2->q1.i );
      cs.s2_[1] = rl_Traits::complex( raydata.s2->q2.r, raydata.s2->q2.i );
      ray.set_polarization( cs );
    }
  }

  return rc;
}

//-------------------------------------------------------------------------
int rot_BPipe::
write( rl_Ray const& ray )
{
  // copy ray/surface data back from ray and surface...

  raydata.pos->x = ray.position(0);
  raydata.pos->y = ray.position(1);
  raydata.pos->z = ray.position(2);

  raydata.dir->x = ray.direction(0);
  raydata.dir->y = ray.direction(1);
  raydata.dir->z = ray.direction(2);

  if ( have_polstate )
  {
    rl_PolCSPOD cs;
    ray.get_PolCSPOD( cs );
    raydata.c2->q1.r = cs.c2_[0].real();
    raydata.c2->q1.i = cs.c2_[0].imag();
    raydata.c2->q2.r = cs.c2_[1].real();
    raydata.c2->q2.i = cs.c2_[1].imag();
    raydata.s2->q1.r = cs.s2_[0].real();
    raydata.s2->q1.i = cs.s2_[0].imag();
    raydata.s2->q2.r = cs.s2_[1].real();
    raydata.s2->q2.i = cs.s2_[1].imag();
  }

  // push the ray into the output pipe
  return bpipe_write_dpkt(bpipe, data, bpo);
}
